// Visuals 
// ==============================================================================
// Visually, the entire sequencer is made up of "GMands" which are meant to read 
// Each mandala is a GPlane + mouse listener + state machine for handling input 
// and transitioning between various modes e.g. hovered, active, playing, etc.
// ==============================================================================

public class GPuz extends GGen {
    // initialize mesh
    .2 => float msize;
    GCube cube --> this;
    cube.scaWorld(@(msize,msize,msize));
    cube.mat().color(@(0,0,0));

    // initialize second hover circle
    GCircle plane --> this;
    plane.scaWorld(@(msize+.1,msize+.1,msize+.1));

    // reference to a mouse
    Mouse @ mouse;
    vec3 mousePos;

    // events
    Event onHoverEvent, onClickEvent;  // onExit, onRelease
    
    // states
    0 => static int NONE;  // not hovered or active
    1 => static int HOVERED;  // hovered
    2 => static int ACTIVE;   // clicked
    3 => static int PLAYING;  // makine sound!
    0 => int state; // current state
    0 => int lastState;

    // input types
    0 => static int MOUSE_HOVER;
    1 => static int MOUSE_EXIT;
    2 => static int MOUSE_CLICK;
    3 => static int NOTE_ON;
    4 => static int NOTE_OFF;

    // color map
    [
        Color.BLACK,    // NONE
        Color.RED,      // HOVERED
        Color.YELLOW,   // ACTIVE
        Color.GREEN     // PLAYING
    ] @=> vec3 colorMap[];

    // NRev reverb => dac;
	BandedWG s => dac;
    float notePos;
    float loopTime;
	0.05 => float vol;
    Math.random2(60,84) => int notes;
    Shred sh;

    0 => s.gain; // set gain
    1 => s.preset; // for BandedWG

    // constructor
    fun void init(Mouse @ m, float noteNum, float notePos, float loopTime) {
        if (mouse != null) return;
        m @=> this.mouse;
    	notePos => this.notePos;  
        loopTime => this.loopTime;
        noteNum$int => notes;                              
        spork ~ playAudio();
        spork ~ pulse() @=> sh;
    }

    fun void color(vec3 c) {
        c => colorMap[NONE];
        c => colorMap[PLAYING];
    }

    // check if state is active (i.e. should play sound)
    fun int active() {
        return state == ACTIVE;
    }

    fun void play() {
        enter(PLAYING);
    }

    // stop play animation (called by sequencer on note off)
    fun void stop() {
        if (state == PLAYING)
            enter(lastState);
        else enter(NONE);
    }

    // enter state, remember last state
    fun void enter(int s) {
        state => lastState;
        s => state;
    }

    fun void decay() {
        while (s.gain() > 0) 
        {
            Math.max(0,s.gain()-0.02) => s.gain;
            200::ms => now;
        }  
    }

    // override ggen update
    fun void update(float dt) {

        // update state
        plane.mat().color(colorMap[state]);
        cube.mat().color( colorMap[state] * .5 + .5*@(1,1,1) );

    }

    fun void collect() {
        this.sca(0.01);
        sh.exit();
        enter(ACTIVE);
        vol* 1.2 => s.gain;
        Std.mtof(notes) => s.freq;
        60 => s.noteOn;
        spork ~ decay();
        play();
    }

    fun void pulse() {
        this.sca() => vec3 origscale;
        origscale.x => float origs;
        while(this.sca().x > .1) {
            while(this.sca().x < origs * 1.1)
            {
                this.sca(this.sca().x + 0.03);
                50::ms => now;
            }
            if( this.posWorld().x > 3 && Math.random2(1,12) == 2) spork ~ flip();
            while(this.sca().x > origs * 0.9)
            {
                this.sca(this.sca().x - 0.03);
                50::ms => now;
            }
            if( this.posWorld().x > 3 && Math.random2(1,12) == 2) spork ~ flip();
        }
    }

    fun void flip() {
        this.rot() => vec3 origrot;
        for(float i; i<2*Math.pi; 0.4 +=> i) {
            this.rotateOnWorldAxis( @(0,1,0), 0.4 );
            50::ms => now;
        }
        this.rot(origrot);
    }

    fun void reset() {
        this.sca(1);
        stop();
        enter(NONE);
    }

	fun void playAudio()
	{
        loopTime::second => dur T;
        T - (now % T) => now; // sync to loop time
        notePos::second => now; // sync to note position
	    while(true)
	    {   
            if (state == PLAYING) {
                vol => s.gain;
                Std.mtof(notes) => s.freq;
                40 => s.noteOn;
	        }
            spork ~ decay();
            T => now;
	    }
	}
}
